-- SEC Engine Script 1.0.2
--   From Ben Laws' 390 script.
--
--CHANGELOG:
--1.0.0 
--  initial release
--1.0.1
--  update coupling configuration - always set a new master script when consist is changed
--  ensure all vehicles are not driving vehicles when not a player consist
--  speed set now works on intervals of 5
--  pivot location variables have been moved to the Update function so they can be classed
--    as global to all functions that use them
-- 1.0.2
--  Modified for RW3 and SEC. DRB 15/9/11
-- 1.0.3
--  BL 21/9/11 Small clean up removing unused legacy RW1 workaround code. 

--Set to false to turn off debug messages
DEBUG = true

function DebugPrint(msg)
	if DEBUG then
		Print(msg)
	end
end


--new stuff:

SEC_MSGID  			= 4901

UID_UNDEFINED 			= -1

COACH_TYPE_UNDEFINED 	= -1
COACH_TYPE_DMRFO 		= 0
COACH_TYPE_DMSO 		= 1
COACH_TYPE_MFO 			= 2
COACH_TYPE_MFOD 		= 3
COACH_TYPE_MSOb 		= 4
COACH_TYPE_MSOd 		= 5
COACH_TYPE_DTPFO 		= 6
COACH_TYPE_DTPSO 		= 7
COACH_TYPE_TSO 			= 8
COACH_TYPE_Reserved1 	= 9
COACH_TYPE_Reserved2 	= 10

POSITION_UPDATE_INTERVAL = 0.05						--seconds

--uid is the unique coach id
_uid = UID_UNDEFINED

--true if this script is the master script
_isMasterScript = false

--a temporary variable for consist message return values
__consistMessageReturnValue = nil

--true if script is initialised
_initialised = false

--type of coach
_coachType = COACH_TYPE_UNDEFINED

--this is true if the last update determined the player as driven
_isPlayerDriven = false
_isEngineWithKey = false

--Neighbour coachs (-1 if no coach is found)
_neighbourCoachForwards = nil
_neighbourCoachBackwards = nil

_currentDoorInterlock = 0
_currentCylinderCockValue = 0
_currentSignalBellValue = 0

_positionUpdateInterval = 0.0

_requiresReInit = false

SynchronizedControls =
{
	["PantographControl"] = true,
	["PantographDirection"] = true,
	["TiltIsolate"] = true
}
--UTILITY FUNCTIONS--
function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

--returns true if lhs and rhs are equal (either nil, or all values are the same)
function coachIsEqual(lhs, rhs)
	if lhs == nil and rhs == nil then return true end
	if lhs == nil or rhs == nil then return false end
	if lhs.uid == rhs.uid and lhs.coachType == rhs.coachType and lhs.isPlayerDriven == rhs.isPlayerDriven then return true end
	return false
end

--returns 1 if the doors are open, 0 if doors are closed
function getDoorInterlockStatus()
	if (GetControlValue("DoorsOpenCloseLeft", 0) > 0.5) or (GetControlValue("DoorsOpenCloseRight", 0) > 0.5) then 
		return 1 
	end
	return 0
end

function isGoingBackwards()
	if (distance(pivotAX, pivotAY, oldPivotBX or pivotAX, oldPivotBY or pivotAY) > distance(oldPivotAX or pivotBX, oldPivotAY or pivotBY, pivotBX, pivotBY)) then
		return true
	else
		return false
	end
end

function string:split(delimiter)
  local result = { }
  local from  = 1
  local index = 1
  local delim_from, delim_to = string.find( self, delimiter, from  )
  while delim_from do
    result[index] = string.sub(self, from , delim_from - 1)
    from  = delim_to + 1
    delim_from, delim_to = string.find( self, delimiter, from  )
	index = index + 1
  end
  result[index] = string.sub(self, from)
  return result
end

function GetControlValue(name, index)
	return Call ("*:GetControlValue", name, index)
end

function GetMinimumControlValue(name, index)
	return Call ("*:GetControlMinimum", name, index)
end

function ControlExists(name, index)
	return Call( "*:ControlExists", name, index ) == 1
end

function IsPlayer()
	return (Call ("GetIsPlayer") == 1)
end

function IsEngineWithKey()
	return (Call ("GetIsEngineWithKey") == 1)
end

function GetSpeed()
	return Call("*:GetSpeed")
end

function SetControlValue(name, index, value)
	Call ("*:SetControlValue", name, index, value)
end

function BeginUpdate()
	Call("BeginUpdate")
end

--END UTILITY FUNCTIONS--


--INTERNALLY CALLED FUNCTIONS

function OnConsistMessage ( msg, argument, direction )
	if msg == SEC_MSGID  then
		local strParts = string.split(argument, ",")
		if strParts[1] == "SlaveRequest" then
			OnSlaveRequest(tonumber(strParts[2]), direction)
		elseif strParts[1] == "EndOfConsist" then
			OnEndOfConsist(tonumber(strParts[2]), direction)
		elseif strParts[1] == "GetNeighbourCoach" then
			OnGetNeighbourCoach(direction)
		elseif strParts[1] == "GetNeighbourCoachResult" then
			OnGetNeighbourCoachResult(tonumber(strParts[2]), tonumber(strParts[3]), tonumber(strParts[4]) == 1, direction)
		elseif strParts[1] == "ConsistChanged" then
			OnConsistChanged(direction)
		elseif strParts[1] == "SetConsistControlValue" then
			OnSetConsistControlValue(strParts[2], tonumber(strParts[3]), tonumber(strParts[4]), direction)
		elseif strParts[1] == "ConsistDoorInterlockRequest" then
			OnConsistDoorInterlockRequest(direction)
		elseif strParts[1] == "ConsistDoorInterlockResponse" then
			OnConsistDoorInterlockResponse(tonumber(strParts[2]), direction)
		elseif strParts[1] == "RequestMasterCoach" then
			OnRequestMasterCoach(direction)
		elseif strParts[1] == "ReturnMasterCoach" then
			OnReturnMasterCoach(tonumber(strParts[2]), tonumber(strParts[3]), tonumber(strParts[4]) == 1, direction)
		end
	end
end

function OnControlValueChange ( name, index, value )
	if _isMasterScript then
		if (name == "CylinderCock") then
			if (value == 1 or value == 0) and value ~= _currentCylinderCockValue then
				_currentCylinderCockValue = value
				if GetControlValue("SpeedSet", 0) == 0 and GetControlValue("TrainBrakeControl", 0) <= 0.01 then
					OnControlValueChange("SpeedSet", 0, 1)
				else
					OnControlValueChange("SpeedSet", 0, 0)
				end
			end
		elseif (name == "Bell") then
			if (value == 1 or value == 0) and value ~= _currentSignalBellValue then
				_currentSignalBellValue = value
				if GetControlValue("SignalBell", 0) == 0 then
					OnControlValueChange("SignalBell", 0, 1)
				else
					OnControlValueChange("SignalBell", 0, 0)
				end
			end
		end
		
		if SynchronizedControls[name] ~= nil then
			SetConsistControlValue(name, index, value, 0)
			SetConsistControlValue(name, index, value, 1)
		end
	end
	SetControlValue(name, index, value)
end

function Initialise()
	_isMasterScript = false
	_initialised = false
	BeginUpdate()
end

function Update(interval)

	--Get the new position
	_positionUpdateInterval = _positionUpdateInterval + interval
	if _positionUpdateInterval >= POSITION_UPDATE_INTERVAL then
		--Get pivot a and b positions
		oldPivotAX = pivotAX
		oldPivotBX = pivotBX
		oldPivotAY = pivotAY
		oldPivotBY = pivotBY
		pivotAX, _, pivotAY = Call("TiltPivotA:getNearPosition")
		pivotBX, _, pivotBY = Call("TiltPivotB:getNearPosition")
		_positionUpdateInterval = 0.0
	end

	--the most important check is to see if the player was driving the consist
	-- and is now not driving it, or was not driving it but is now driving it:
	if _isPlayerDriven ~= IsPlayer() or _isEngineWithKey ~= IsEngineWithKey() then
		_initialised = false
		_isMasterScript = false
		_requiresReInit = false --just incase a ConsistChanged message set it to true
		_uid = UID_UNDEFINED
		if DEBUG then
			if _isPlayerDriven ~= IsPlayer() then
				DebugPrint("This coach has uninitialised because IsPlayer() is different to the stored value")
			end
		end
		_isPlayerDriven = IsPlayer()
		_isEngineWithKey = IsEngineWithKey()
		--note that this will cause any vehicle that was once a separate consist to become uninitialised
	end

	
	-- detect consist changes - bearing in mind that at any time a consist could change
	--this must be done before the script updates anything else
	if _initialised then
		local newNeighbourCoachForwards = GetNeighbourCoach(0)
		local newNeighbourCoachBackwards = GetNeighbourCoach(1)
		
		if not (coachIsEqual(newNeighbourCoachForwards, _neighbourCoachForwards) and coachIsEqual(newNeighbourCoachBackwards, _neighbourCoachBackwards) ) then
			--consist has changed
			
			_initialised = false
			_uid = UID_UNDEFINED
			if _isMasterScript then
				_requiresReInit = true
			else
				SendConsistChangedMessage(0)
				SendConsistChangedMessage(1)
			end
		end
	end
	
	if _initialised then
		--Standalone functions--
		
		Tilt(interval)
		PantoArcing(interval)
		PantographArmAnimation(interval)
		
		--only the master script needs to manage this stuff
		if _isMasterScript then
			_currentDoorInterlock = getDoorInterlockStatus()
			ConsistDoorInterlockRequest()
	
			--Standalone functions--
			TractionSound(interval)
			CombinedThrottleAndBrake(interval)
		end
	else
		
		--first thing to check is if this is player driven
		--if it is not player driven, set it as master
		--if it is player driven, check that this is the driving car
		
		--the master script can be any vehicle if it is not player driven
		--however, the script will work best when the master script is the player's vehicle
		if _requiresReInit or not IsPlayer() or IsEngineWithKey() then
			_requiresReInit = false
			--if it is, then don't initialise unless this is the driving car
			
			setAsMaster()
			setupPlayerVehicle()
		end
		
	end
end

--this is a common initialisation function
function commonInit(id, setMaster)
	_isMasterScript = setMaster
	_uid = id
	_coachType = round(GetMinimumControlValue("CoachID", 0))
	_initialised = true
	_isPlayerDriven = IsPlayer()
	_isEngineWithKey = IsEngineWithKey()
	
	if DEBUG then
		local debugString = "Initialised with parameters. _uid: " .. _uid
		if (_isMasterScript) then
			debugString = debugString .. ", SET AS MASTER"
		else
			debugString = debugString .. ", SET AS SLAVE"
		end
	
		if _isPlayerDriven then
			debugString = debugString .. ", Player driven"
		else
			debugString = debugString .. ", AI"
		end
		DebugPrint(debugString)
	
	end
end

function setAsMaster()
	commonInit(0, true)
	local nextID = SlaveRequest(1, 0)
	local idCount = SlaveRequest(nextID, 1)
	_neighbourCoachForwards = GetNeighbourCoach(0)
	_neighbourCoachBackwards = GetNeighbourCoach(1)
end

function createCoachTable(uid, coachType, isPlayerDriven)
	local coachTable = { ["uid"] = uid, ["coachType"] = coachType, ["isPlayerDriven"] = isPlayerDriven }
	return coachTable
end

function GetNeighbourCoach(direction)
	__consistMessageReturnValue = nil
	Call("SendConsistMessage", SEC_MSGID , "GetNeighbourCoach", direction)
	return __consistMessageReturnValue
end

function GetNeighbourCoachResult(direction, uid, coachType, isPlayerDriven)
	local boolToNum
	if isPlayerDriven then boolToNum = 1 else boolToNum = 0 end
	Call("SendConsistMessage", SEC_MSGID , "GetNeighbourCoachResult," .. uid .. "," .. coachType .. "," .. boolToNum, direction)
end

--send a slave request
function SlaveRequest(id, direction)
	__consistMessageReturnValue = id
	Call("SendConsistMessage", SEC_MSGID , "SlaveRequest," .. id, direction)
	return __consistMessageReturnValue
end

function ConsistDoorInterlockRequest(direction)
	if direction == nil then
		Call("SendConsistMessage", SEC_MSGID , "ConsistDoorInterlockRequest", 0)
		Call("SendConsistMessage", SEC_MSGID , "ConsistDoorInterlockRequest", 1)
	else
		Call("SendConsistMessage", SEC_MSGID , "ConsistDoorInterlockRequest", direction)
	end
end

function ConsistDoorInterlockResponse(interlockStatus, direction)
	Call("SendConsistMessage", SEC_MSGID , "ConsistDoorInterlockResponse," .. interlockStatus, direction)
end

--send the message that the consist has ended
function EndOfConsist(id, direction)
	Call("SendConsistMessage", SEC_MSGID , "EndOfConsist," .. id, direction)
end

function SendConsistChangedMessage(direction)
	Call("SendConsistMessage", SEC_MSGID , "ConsistChanged", direction)
end

function SetConsistControlValue(name, index, value, direction)
	local msg = "SetConsistControlValue," .. name .. "," .. index .. "," .. value
	if direction == nil then
		Call("SendConsistMessage", SEC_MSGID , msg, 0)
		Call("SendConsistMessage", SEC_MSGID , msg, 1)
	else
		Call("SendConsistMessage", SEC_MSGID , msg, direction)
	end
end

function GetMasterCoach()
	if isMasterCoach then
		return createCoachTable(_uid, _coachType, _isPlayerDriven)
	end
	__consistMessageReturnValue = nil
	RequestMasterCoach(0)
	if __consistMessageReturnValue == nil then
		RequestMasterCoach(1)
	end
	return __consistMessageReturnValue
end

function RequestMasterCoach(direction)
	Call("SendConsistMessage", SEC_MSGID , "RequestMasterCoach", direction)
end

function ReturnMasterCoach(uid, coachType, isPlayerDriven, direction)
	local boolToNum
	if isPlayerDriven then boolToNum = 1 else boolToNum = 0 end
	Call("SendConsistMessage", SEC_MSGID , "ReturnMasterCoach," .. uid .. "," .. coachType .. "," .. boolToNum, direction)
end

--This is called as a result of the simulation script
--being setup
function setupPlayerVehicle()
	SetControlValue ("PantographDirection", 0, 1)
	SetConsistControlValue ("PantographDirection", 0, 1)
	SetControlValue ("PantographControl", 0, 1)
	SetConsistControlValue ("PantographControl", 0, 1)
	SetControlValue ("TiltIsolate", 0, 0)
	SetConsistControlValue ("TiltIsolate", 0, 0)
end
--RECEIVED CONSIST MESSAGES--

--receive a slave request
function OnSlaveRequest(id, direction)
	commonInit(id, false)
	
	id = id + 1
	local nextID = SlaveRequest(id, direction)
	_neighbourCoachForwards = GetNeighbourCoach(0)
	_neighbourCoachBackwards = GetNeighbourCoach(1)
	--if the next id to use is the same as the id sent, then the end of the
	--consist was found
	if nextID == id then
		EndOfConsist(id, 1 - direction)
	end
end

function OnEndOfConsist(id, direction)
	__consistMessageReturnValue = id
	if not _isMasterScript then
		EndOfConsist(id, direction)
	end
end

function OnGetNeighbourCoach(direction)
	GetNeighbourCoachResult(1 - direction, _uid, _coachType, _isPlayerDriven)
end

function OnGetNeighbourCoachResult(uid, coachType, isPlayerDriven, direction)
	__consistMessageReturnValue = createCoachTable(uid, coachType, isPlayerDriven)
end

function OnConsistChanged(direction)
	_initialised = false
	_uid = UID_UNDEFINED
	if _isMasterScript then
		_requiresReInit = true
	else
		SendConsistChangedMessage(direction)
	end
end

function OnSetConsistControlValue(name, index, value, direction)
	SetControlValue(name, index, value)
	SetConsistControlValue(name, index, value, direction)
end

function OnConsistDoorInterlockRequest(direction)
	ConsistDoorInterlockResponse(getDoorInterlockStatus(), 1 - direction)
	ConsistDoorInterlockRequest(direction)
end

function OnConsistDoorInterlockResponse(interlock, direction)
	if _isMasterScript then
		_currentDoorInterlock = math.max(interlock, _currentDoorInterlock)
	else
		ConsistDoorInterlockResponse(interlock, direction)
	end
end

function OnRequestMasterCoach(direction)
	if _isMasterScript then
		ReturnMasterCoach(_uid, _coachType, _isPlayerDriven, 1 - direction)
	else
		RequestMasterCoach(direction)
	end
end

function OnReturnMasterCoach(uid, coachType, isPlayerDriven, direction)
	__consistMessageReturnValue = createCoachTable(uid, coachType, isPlayerDriven)
	ReturnMasterCoach(uid, coachType, isPlayerDriven, direction)
end
--END RECEIVED CONSIST MESSAGES--





--STANDALONE FUNCTIONS

----------------------------------TRACTION SOUND

-- Traction Sound Variables --
curTractionSoundTime = 0.0  -- seconds
maxTractionSoundTime = 20.0 -- seconds
tractionSoundStarting = false
------------------------------

function TractionSound(interval)
	--Traction sound:
	--  get the pantograph control value
	
	--  get the regulator control value, this is actually calculated from ThrottleAndBrake
	--  regulator will be set automatically to 0 on emerg. brake, however ThrottleAndBrake
	--  should override this
	local speed = math.abs(GetSpeed())
	local panto = GetControlValue("PantographControl", 0)
	local value
	if IsPlayer() then
		value = GetControlValue("Regulator", 0) * panto
	else
		value = panto * (speed - 0.1)
	end
	-- must be moving: --EDIT not any more
	--if speed1 == 0 and speed2 == 0 and speed3 == 0 then value = 0 else value = panto end
	
	--  find out if traction sound is already on, does not count if TractionSound is less than 1 but greater than 0
	local tractionOn = GetControlValue("TractionSound", 0) == 1
	
	if value > 0.0 then
		-- if power is being applied
		if not tractionOn then
			tractionSoundStarting = true
		end
	else
		-- no power applied, traction sound on
		if tractionOn then
			-- make sure speed is about 0, and begin fan shutdown sound after some time
			if speed < 0.05 then
				curTractionSoundTime = curTractionSoundTime + interval
			else
				curTractionSoundTime = 0.0
			end
			
			-- if panto down, not enough power for fans, so shut them down
			if panto == 0 then curTractionSoundTime = maxTractionSoundTime end
			
			if (curTractionSoundTime >= maxTractionSoundTime) then
				-- turn of sound
				SetControlValue("TractionSound", 0, 0)
				SetConsistControlValue("TractionSound", 0, 0)
			end
		end
	end
	
	if tractionSoundStarting then
		-- increase TractionSound gradually using interval, so it takes 1 second to reach 1
		local tsound = GetControlValue("TractionSound", 0) + interval
		if tsound >= 1 then 
			SetControlValue("TractionSound", 0, 1) -- max TractionSound to 1
			SetConsistControlValue("TractionSound", 0, 1) -- max TractionSound to 1
			tractionSoundStarting = false
		else 
			SetControlValue("TractionSound", 0, tsound) 
			SetConsistControlValue("TractionSound", 0, tsound) 
		end
		curTractionSoundTime = 0.0
	end
end


----------------------------------THROTTLE AND BRAKE
--includes speed set algorithms

MAX_TRACTIVE_EFFORT = 22.7 	--kN
DOORS_CLOSING_TIME = 10 	--seconds
SIGNAL_BELL_TIME_1 = 2 		--seconds
SIGNAL_BELL_TIME_2 = 2.8 	--seconds
DYNAMIC_BRAKE_VALUES = { [0] = 0, [22] = 0, [25] = 1.0, [65] = 1.0, [92] = 0.75, [130] = 0.5, [160] = 0.5 }
MPH_TO_MS = 0.44704 		--value to multiply mph by to get m/s
SPEED_SET_MPHINTERVAL = 5	-- mph

regulatorRequiresReset = false
doorsInterlocked = false
doorInterlockTimer = 0.0
waitingOnSignalBell1 = false
waitingOnSignalBell2 = false
signalBellTimer = 0.0
oldCombinedThrottleBrake = 0
speedJustSet = false
speedSetVelocity = 0.0
speedSetUpdateTime = 0.1
speedSetLastTime = 0.0

function CombinedThrottleAndBrake(interval)
	--does not apply to AI
	--just set the motor sound, that's all
	if not IsPlayer() then 
		SetControlValue("MotorSound", 0, 0.5)
		SetConsistControlValue("MotorSound", 0, 0.5)
		return
	end
	
	-- get emergency brake value. Problem here: this does not apply to consist when it's automatic
	local eb = GetControlValue("EmergencyBrake", 0)
	local rev = GetControlValue("Reverser", 0)
	local speed = math.abs(GetSpeed())
	
	--if emergency brake is applied or the reverser is set to off then
	-- the regulator requires a reset. Regulator is reset by
	-- putting the combined throttle / brake into off position

	if eb > 0 or rev == 0 then
		regulatorRequiresReset = true
	end
	local tAndBVal = GetControlValue("ThrottleAndBrake", 0)
	if (tAndBVal ~= nil) then
		local brakeVal = 0
		local throttleVal = 0
		if (tAndBVal <= 0) then
			regulatorRequiresReset = false
			brakeVal = -tAndBVal
		else
			throttleVal = tAndBVal * 2
		end

		--First, update the interlock status stuff
		
		if _currentDoorInterlock == 1 then
			SetControlValue("DoorsInterlockLeft", 0, 1)
			SetControlValue("DoorsInterlockRight", 0, 1)
			regulatorRequiresReset = true
			doorsInterlocked = true
			doorInterlockTimer = 0.0
		else
			if (doorsInterlocked) then
				doorInterlockTimer = doorInterlockTimer + interval
				if (doorInterlockTimer >= DOORS_CLOSING_TIME) then 
					doorsInterlocked = false 
					SetControlValue("DoorsInterlockLeft", 0, 0)
					SetControlValue("DoorsInterlockRight", 0, 0)
					waitingOnSignalBell1 = true
					signalBellTimer = 0.0
				end
			end
		end
		
		--use the signal bell
		
		if waitingOnSignalBell1 then
			signalBellTimer = signalBellTimer + interval
			if signalBellTimer >= SIGNAL_BELL_TIME_1 then
				local sbVal = 1 - GetControlValue("SignalBell", 0)
				SetControlValue("SignalBell", 0, sbVal)
				waitingOnSignalBell1 = false
				waitingOnSignalBell2 = true
			end
		elseif waitingOnSignalBell2 then
			signalBellTimer = signalBellTimer + interval
			if signalBellTimer >= SIGNAL_BELL_TIME_2 then
				local sbVal = 1 - GetControlValue("SignalBell", 0)
				SetControlValue("SignalBell", 0, sbVal)
				waitingOnSignalBell2 = false
			end
		end

		-- dont use brakeVal if door interlock is on
		if (doorsInterlocked and brakeVal < 0.76) then
			brakeVal = 0.76
		end

		local percentDynamic = 0
		
		--only emulates dynamic brake
		if brakeVal < 1 then
			local mph = speed / MPH_TO_MS
			local finalMph = nil
			local lastMph = 0
			local finalPercent = 0
			for k,v in pairs(DYNAMIC_BRAKE_VALUES) do
				if (k > mph) then
					if (finalMph == nil or k - mph < finalMph - mph) then
						finalMph = k
						finalPercent = v
					end
				else
					if (mph - k < mph - lastMph) then
						lastMph = k
						percentDynamic = v
					end
				end
			end
			if finalMph ~= nil then
				mphRange = finalMph - lastMph
				percentRange = finalPercent - percentDynamic
				percentDynamic = (((mph - lastMph) / mphRange) * percentRange + percentDynamic)
			end
			if (percentDynamic == nil) then percentDynamic = 0 end
		end
		SetControlValue("TrainBrakeControl", 0, brakeVal)
		SetConsistControlValue("TrainBrakeControl", 0, brakeVal)
		local accel = math.min(brakeVal * percentDynamic * MAX_TRACTIVE_EFFORT, 99)
		local accelTens = math.floor(accel / 10)
		local accelUnits = accel - (accelTens * 10)
		SetControlValue("DynaBrakeEffortUnits", 0, accelUnits)
		SetControlValue("DynaBrakeEffortTens", 0, accelTens)
		local tbcpValue = GetControlValue("TrainBrakeCylinderPressureBAR", 0)
		if (tbcpValue ~= nil) then
			tbcpValue = tbcpValue * (1 - percentDynamic)
			SetControlValue("TbcpNewBAR", 0, tbcpValue)
		end

		if brakeVal > 0 or regulatorRequiresReset then
			throttleVal = 0
		else
			local speedSet = GetControlValue("SpeedSet", 0) == 1
			if (speedSet) then
				if (not speedJustSet) then
					speedSetVelocity = math.floor((speed / MPH_TO_MS) / SPEED_SET_MPHINTERVAL) * SPEED_SET_MPHINTERVAL * MPH_TO_MS
					speedJustSet = true
				end
				local maxThrottle = throttleVal
				throttleVal = GetControlValue("Regulator", 0)
				speedSetLastTime = speedSetLastTime + interval
				if (speedSetLastTime >= speedSetUpdateTime) then
					if (speedSetVelocity < speed) then
						throttleVal = math.min(throttleVal - 0.3 * speedSetLastTime, maxThrottle)
					elseif (speedSetVelocity > speed) then
						throttleVal = math.min(throttleVal + 0.3 * speedSetLastTime, maxThrottle)
					end
					speedSetLastTime = 0.0
				end
			else
				speedJustSet = false
				speedSetVelocity = 0
			end
		end
		SetControlValue("Regulator", 0, throttleVal)
		SetConsistControlValue("Regulator", 0, throttleVal)
		SetControlValue("MotorSound", 0, math.max(throttleVal, brakeVal))
		SetConsistControlValue("MotorSound", 0, math.max(throttleVal, brakeVal))
	end

	--local throttleBrake = throttleAndBrake
	if (tAndBVal ~= oldCombinedThrottleBrake or eb > 0 or regulatorRequiresReset) then SetControlValue("SpeedSet", 0, 0) end
	oldCombinedThrottleBrake = tAndBVal
end




------------------------------- PANTOGRAPH ARM ANIMATION
PANTO_ARM_RAISE_TIME_MULTIPLIER = 1.5

function PantographArmAnimation(interval)

	if _coachType == COACH_TYPE_DTPFO or _coachType == COACH_TYPE_DTPSO then

		local raised

		if IsPlayer() then

-----------------------------------------------------------------------------------
-- Following replaces 390 script for RW3. DRB 15/9/11

			if not _isEngineWithKey and GetControlValue("PantographControl", 0) == 1 then
				raised = 1
			else
				raised = 0
			end

		else

			if GetSpeed() < 0 and GetControlValue("PantographControl", 0) == 1 then 
				raised = 1
			else
				raised = 0
			end

		end

-----------------------------------------------------------------------------------

		SetControlValue("PantographArm", 0, raised)

		if GetControlValue("PantographArm", 0) then
			if GetControlValue("PantographArm", 0) > 0.5 then
				Call("*:AddTime", "panto", interval * PANTO_ARM_RAISE_TIME_MULTIPLIER)
			else
				Call("*:AddTime", "panto", -interval * PANTO_ARM_RAISE_TIME_MULTIPLIER)
			end
		end
	end
end



--- START ARCING SPECIFIC ---

ARCING_MAX_CHANCE = 0.995
ARCING_MIN_CHANCE = 0.25
INCREASE_CHANCE = 0.35
DECREASE_CHANCE = 0.1
TIME_UNTIL_PANTO_RAISE_ARC = 0.5 --second
arcingCurrentChance = ARCING_MAX_CHANCE
lastSparkVal = -1 --use this to ensure sparks always look different if repeated in consecutive frames
pantoArmRaisingTime = 0.0

function generateSpark(useSound, value)
	local sparkVal = value or math.floor(13 * math.random())
	if (sparkVal == lastSparkVal) then
		if (sparkVal == 13) then sparkVal = 0
		else sparkVal = sparkVal + 1 end
	end
	lastSparkVal = sparkVal
	if (sparkVal == 0) then
		Call("*:ActivateNode", "sparkA", 1)
		Call("*:ActivateNode", "sparkC", 1)
		Call("*:ActivateNode", "sparkE", 1)
	elseif (sparkVal == 1) then
		Call("*:ActivateNode", "sparkB", 1)
		Call("*:ActivateNode", "sparkD", 1)
		Call("*:ActivateNode", "sparkF", 1)
	elseif (sparkVal == 2) then
		Call("*:ActivateNode", "sparkC", 1)
		Call("*:ActivateNode", "sparkE", 1)
	elseif (sparkVal == 3) then
		Call("*:ActivateNode", "sparkD", 1)
		Call("*:ActivateNode", "sparkF", 1)
	elseif (sparkVal == 4) then
		Call("*:ActivateNode", "sparkE", 1)
	elseif (sparkVal == 5) then
		Call("*:ActivateNode", "sparkF", 1)
	elseif (sparkVal == 6) then
		Call("*:ActivateNode", "sparkA", 1)
		Call("*:ActivateNode", "sparkE", 1)
	elseif (sparkVal == 7) then
		Call("*:ActivateNode", "sparkB", 1)
		Call("*:ActivateNode", "sparkF", 1)
	elseif (sparkVal == 8) then
		Call("*:ActivateNode", "sparkA", 1)
		Call("*:ActivateNode", "sparkC", 1)
	elseif (sparkVal == 9) then
		Call("*:ActivateNode", "sparkB", 1)
		Call("*:ActivateNode", "sparkD", 1)
	elseif (sparkVal == 10) then
		Call("*:ActivateNode", "sparkA", 1)
	elseif (sparkVal == 11) then
		Call("*:ActivateNode", "sparkB", 1)
	elseif (sparkVal == 12) then
		Call("*:ActivateNode", "sparkC", 1)
	elseif (sparkVal == 13) then
		Call("*:ActivateNode", "sparkD", 1)
	end
	if (useSound) then
		if (GetControlValue("Spark", 0) == 0) then
			SetControlValue("Spark", 0, 1)
		else
			SetControlValue("Spark", 0, 0)
		end
	end
end

function PantoArcing(interval)
	--lastSparkVal = -1
	local speed = math.abs(GetSpeed())
	local activePanto = GetControlValue("PantographArm", 0)
	if activePanto then
		Call("*:ActivateNode", "sparkA", 0)
		Call("*:ActivateNode", "sparkB", 0)
		Call("*:ActivateNode", "sparkC", 0)
		Call("*:ActivateNode", "sparkD", 0)
		Call("*:ActivateNode", "sparkE", 0)
		Call("*:ActivateNode", "sparkF", 0)
		if (activePanto == 1) then
			if (pantoArmRaisingTime < TIME_UNTIL_PANTO_RAISE_ARC) then
				pantoArmRaisingTime = pantoArmRaisingTime + interval
				if pantoArmRaisingTime >= TIME_UNTIL_PANTO_RAISE_ARC then
					generateSpark(false)
				end
			end
			if (speed > 0.1) then
				if (arcingCurrentChance < ARCING_MAX_CHANCE) then 
					arcingCurrentChance = math.min(arcingCurrentChance + DECREASE_CHANCE, ARCING_MAX_CHANCE) 
				end
				if (math.random() > arcingCurrentChance) then
					arcingCurrentChance = arcingCurrentChance - INCREASE_CHANCE --increase the chance of more sparks at same time
					if (arcingCurrentChance < ARCING_MIN_CHANCE) then arcingCurrentChance = ARCING_MIN_CHANCE end
					--pick a random spark
					generateSpark(true)
				
				end
			end
		else
			if (pantoArmRaisingTime > 0.0) then
				generateSpark(false)
				pantoArmRaisingTime = 0.0
			end
		end
	end
end

--- END ARCING SPECIFIC ---






--------------------------------------- TILT

-- Tilt Variables --
TILT_FRAME_COUNT = 17.0							--frames
TILT_CENTRE_FRAME = (TILT_FRAME_COUNT + 1) / 2	--frames
TILT_MAX_DEGREES = (8.0 * math.pi) / 180.0		--rads
TILT_FPS = 31.0		  							--frames per sec
TILT_CURVATURE_TO_FRAME_FACTOR = 100  			--multiply the heading angle by this to get the frame difference from centre
TILT_INTERVAL_FACTOR = 3 						--multiply the update interval by this to update tilt animation								
TILT_UPDATE_INTERVAL = 0.05						--seconds
TILT_CUT_OFF_SPEED = 20							--metres per second
TILT_ACCELERATION = 6
TILT_DECELERATION = 1.5
tiltInterval = 0.0								--cumulative interval for updating tilt calculations
curTiltSpeed = 0.0
curFrame = TILT_CENTRE_FRAME					--current tilt frame
newFrame = TILT_CENTRE_FRAME					--target tilt frame
----------------------------------------

--Calculate / perform tilt
function Tilt(interval)
	local speed = math.abs(GetSpeed())
	tiltInterval = tiltInterval + interval
	if (tiltInterval >= TILT_UPDATE_INTERVAL) then
		--push down old values

		local a1 = a2 or calcHeading(pivotBX, pivotBY, pivotAX, pivotAY)
		a2 = calcHeading(pivotBX, pivotBY, pivotAX, pivotAY)
		local angle = (a2 - a1)
		local tiltAmount
		
		if (GetControlValue("TiltIsolate", 0) == 1 or GetControlValue("PantographControl", 0) == 0) then
			tiltAmount = 0
		else
			tiltAmount = math.abs(Call("*:GetCurvature") * math.max(speed - TILT_CUT_OFF_SPEED, 0))
		end
		
		tiltInterval = 0.0

		--work out direction
		if isGoingBackwards() then
			sign = 1
		else
			sign = -1
		end
		
		if (angle == 0) then sign = 0
		elseif (angle < 0) then sign = -sign end
		newFrame = math.min(math.max(TILT_CENTRE_FRAME + sign * (tiltAmount * TILT_CURVATURE_TO_FRAME_FACTOR), 1), TILT_FRAME_COUNT)
		
	end
	tiltTowards(interval)
end

--Calculates distance between xa, ya and xb, yb
function distance(xa, ya, xb, yb)
	if xa and ya and xb and yb then
		return math.abs(round(math.sqrt(math.pow(xb - xa, 2) + math.pow(yb - ya, 2))))
	else
		return 0
	end
end

--Calculates the atan2 angle between x1, y1 and x2, y2
function calcHeading(x1, y1, x2, y2)
	local angle = math.atan2((y2 - y1),(x2 - x1))
	return angle
end

--Updates tilt animation
function tiltTowards(interval)
	if (newFrame < curFrame) then
		curTiltSpeed = math.max(curTiltSpeed - (TILT_ACCELERATION * interval), -TILT_INTERVAL_FACTOR)
		curTiltSpeed = math.max(curTiltSpeed, (newFrame - curFrame) * TILT_DECELERATION)
	elseif (newFrame > curFrame) then
		curTiltSpeed = math.min(curTiltSpeed + (TILT_ACCELERATION * interval), TILT_INTERVAL_FACTOR)
		curTiltSpeed = math.min(curTiltSpeed, (newFrame - curFrame) * TILT_DECELERATION)
	else
		curTiltSpeed = 0
	end
	
	curFrame = curFrame + interval * curTiltSpeed

--[[
	if (newFrame < curFrame) then
		curFrame = curFrame - math.min(interval * TILT_INTERVAL_FACTOR, interval * (curFrame - newFrame))
		if (curFrame < newFrame) then curFrame = newFrame end
	elseif (newFrame > curFrame) then
		curFrame = curFrame + math.min(interval * TILT_INTERVAL_FACTOR, interval * (newFrame - curFrame))
		if (curFrame > newFrame) then curFrame = newFrame end
	end
	]]
	Call("*:Reset", "tilt")
	Call("*:AddTime", "tilt", curFrame / TILT_FPS)
	SetControlValue("TiltTime", 0, curFrame)
end
